==> aria2-1.1.2.tar.bz2: aria2-1.1.2/src/RarestPieceSelector.cc <==
/* <!-- copyright */
/*
 * aria2 - The high speed download utility
 *
 * Copyright (C) 2006 Tatsuhiro Tsujikawa
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */
/* copyright --> */
#include "RarestPieceSelector.h"

#include <algorithm>

#include "SimpleRandomizer.h"

namespace aria2 {

PieceStat::PieceStat(size_t index):_order(0), _index(index), _count(0) {}

bool PieceStat::operator<(const PieceStat& pieceStat) const
{
  if(_count == pieceStat._count) {
    return _order < pieceStat._order;
  } else {
    return _count < pieceStat._count;
  }
}

void PieceStat::addCount()
{
  if(_count < SIZE_MAX) {
    ++_count;
  }
}

void PieceStat::subCount()
{
  if(_count > 0) {
    --_count;
  }
}

size_t PieceStat::getIndex() const
{
  return _index;
}

size_t PieceStat::getCount() const
{
  return _count;
}

void PieceStat::setOrder(size_t order)
{
  _order = order;
}

size_t PieceStat::getOrder() const
{
  return _order;
}

class GenPieceStat {
private:
  size_t _index;
public:
  GenPieceStat():_index(0) {}

  SharedHandle<PieceStat> operator()()
  {
    return SharedHandle<PieceStat>(new PieceStat(_index++));
  }
};

RarestPieceSelector::RarestPieceSelector(size_t pieceNum, bool randomShuffle):
  _pieceStats(pieceNum)
{
  std::generate(_pieceStats.begin(), _pieceStats.end(), GenPieceStat());
  _sortedPieceStats = _pieceStats;
  // we need some randomness in ordering.
  if(randomShuffle) {
    std::random_shuffle(_sortedPieceStats.begin(), _sortedPieceStats.end(),
			*(SimpleRandomizer::getInstance().get()));
  }
  {
    size_t order = 0;
    for(std::deque<SharedHandle<PieceStat> >::iterator i = _sortedPieceStats.begin();
	i != _sortedPieceStats.end(); ++i) {
      (*i)->setOrder(order++);
    }
  }
}

class FindRarestPiece
{
private:
  const std::deque<size_t>& _indexes;
public:
  FindRarestPiece(const std::deque<size_t>& indexes):_indexes(indexes) {}

  bool operator()(const SharedHandle<PieceStat>& pieceStat)
  {
    return std::binary_search(_indexes.begin(), _indexes.end(), pieceStat->getIndex());
  }
};

bool RarestPieceSelector::select
(size_t& index,
 const std::deque<size_t>& candidateIndexes) const
{
  std::deque<SharedHandle<PieceStat> >::const_iterator i =
    std::find_if(_sortedPieceStats.begin(), _sortedPieceStats.end(),
		 FindRarestPiece(candidateIndexes));
  if(i == _sortedPieceStats.end()) {
    return false;
  } else {
    index = (*i)->getIndex();
    return true;
  }
}

void RarestPieceSelector::addPieceStats(const unsigned char* bitfield,
					size_t bitfieldLength)
{
  size_t index = 0;
  for(size_t bi = 0; bi < bitfieldLength; ++bi) {
    
    for(size_t i = 0; i < 8; ++i, ++index) {
      unsigned char mask = 128 >> i;
      if(bitfield[bi]&mask) {
	_pieceStats[index]->addCount();
      }
    }

  }
  std::sort(_sortedPieceStats.begin(), _sortedPieceStats.end());
}

void RarestPieceSelector::subtractPieceStats(const unsigned char* bitfield,
					     size_t bitfieldLength)
{
  size_t index = 0;
  for(size_t bi = 0; bi < bitfieldLength; ++bi) {
    
    for(size_t i = 0; i < 8; ++i, ++index) {
      unsigned char mask = 128 >> i;
      if(bitfield[bi]&mask) {
	_pieceStats[index]->subCount();
      }
    }

  }
  std::sort(_sortedPieceStats.begin(), _sortedPieceStats.end());
}

void RarestPieceSelector::updatePieceStats(const unsigned char* newBitfield,
					   size_t newBitfieldLength,
					   const unsigned char* oldBitfield)
{
  size_t index = 0;
  for(size_t bi = 0; bi < newBitfieldLength; ++bi) {
    
    for(size_t i = 0; i < 8; ++i, ++index) {
      unsigned char mask = 128 >> i;
      if((newBitfield[bi]&mask) && !(oldBitfield[bi]&mask)) {
	_pieceStats[index]->addCount();
      } else if(!(newBitfield[bi]&mask) && (oldBitfield[bi]&mask)) {
	_pieceStats[index]->subCount();
      }
    }

  }
  std::sort(_sortedPieceStats.begin(), _sortedPieceStats.end());
}

void RarestPieceSelector::addPieceStats(size_t index)
{
  SharedHandle<PieceStat> pieceStat(_pieceStats[index]);
  {
    std::deque<SharedHandle<PieceStat> >::iterator cur =
      std::lower_bound(_sortedPieceStats.begin(), _sortedPieceStats.end(),
		       pieceStat);
    _sortedPieceStats.erase(cur);
  }
  pieceStat->addCount();

  std::deque<SharedHandle<PieceStat> >::iterator to =
    std::lower_bound(_sortedPieceStats.begin(), _sortedPieceStats.end(),
		     pieceStat);

  _sortedPieceStats.insert(to, pieceStat);
}

const std::deque<SharedHandle<PieceStat> >&
RarestPieceSelector::getSortedPieceStats() const
{
  return _sortedPieceStats;
}

} // namespace aria2

==> aria2-1.3.1.tar.bz2: aria2-1.3.1/src/RarestPieceSelector.cc <==
/* <!-- copyright */
/*
 * aria2 - The high speed download utility
 *
 * Copyright (C) 2006 Tatsuhiro Tsujikawa
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */
/* copyright --> */
#include "RarestPieceSelector.h"

#include <cassert>
#include <algorithm>

#include "SimpleRandomizer.h"

namespace aria2 {

PieceStat::PieceStat(size_t index):_order(0), _index(index), _count(0) {}

bool PieceStat::operator<(const PieceStat& pieceStat) const
{
  if(_count == pieceStat._count) {
    return _order < pieceStat._order;
  } else {
    return _count < pieceStat._count;
  }
}

void PieceStat::addCount()
{
  if(_count < SIZE_MAX) {
    ++_count;
  }
}

void PieceStat::subCount()
{
  if(_count > 0) {
    --_count;
  }
}

size_t PieceStat::getIndex() const
{
  return _index;
}

size_t PieceStat::getCount() const
{
  return _count;
}

void PieceStat::setOrder(size_t order)
{
  _order = order;
}

size_t PieceStat::getOrder() const
{
  return _order;
}

class GenPieceStat {
private:
  size_t _index;
public:
  GenPieceStat():_index(0) {}

  SharedHandle<PieceStat> operator()()
  {
    return SharedHandle<PieceStat>(new PieceStat(_index++));
  }
};

RarestPieceSelector::RarestPieceSelector(size_t pieceNum, bool randomShuffle):
  _pieceStats(pieceNum),
  _sortedPieceStatIndexes(pieceNum)
{
  std::generate(_pieceStats.begin(), _pieceStats.end(), GenPieceStat());
  std::vector<SharedHandle<PieceStat> > sortedPieceStats(_pieceStats);
  // we need some randomness in ordering.
  if(randomShuffle) {
    std::random_shuffle(sortedPieceStats.begin(), sortedPieceStats.end(),
			*(SimpleRandomizer::getInstance().get()));
  }
  {
    size_t order = 0;
    for(std::vector<SharedHandle<PieceStat> >::iterator i =
	  sortedPieceStats.begin(); i != sortedPieceStats.end(); ++i) {
      _sortedPieceStatIndexes[order] = (*i)->getIndex();
      (*i)->setOrder(order++);
    }
  }  
}

class FindRarestPiece
{
private:
  const unsigned char* _misbitfield;
  size_t _numbits;
public:
  FindRarestPiece(const unsigned char* misbitfield, size_t numbits):
    _misbitfield(misbitfield), _numbits(numbits) {}

  bool operator()(const size_t& index)
  {
    assert(index < _numbits);
    unsigned char mask = (128 >> (index%8));
    return _misbitfield[index/8]&mask;
  }
};

bool RarestPieceSelector::select
(size_t& index, const unsigned char* bitfield, size_t nbits) const
{
  std::vector<size_t>::const_iterator i =
    std::find_if(_sortedPieceStatIndexes.begin(), _sortedPieceStatIndexes.end(),
		 FindRarestPiece(bitfield, nbits));
  if(i == _sortedPieceStatIndexes.end()) {
    return false;
  } else {
    index = *i;
    return true;
  }
}

class PieceStatRarer {
private:
  const std::vector<SharedHandle<PieceStat> >& _pieceStats;
public:
  PieceStatRarer(const std::vector<SharedHandle<PieceStat> >& ps):
    _pieceStats(ps) {}

  bool operator()(size_t lhs, size_t rhs) const
  {
    return _pieceStats[lhs] < _pieceStats[rhs];
  }
};

void RarestPieceSelector::addPieceStats(const unsigned char* bitfield,
					size_t bitfieldLength)
{
  size_t index = 0;
  for(size_t bi = 0; bi < bitfieldLength; ++bi) {
    
    for(size_t i = 0; i < 8; ++i, ++index) {
      unsigned char mask = 128 >> i;
      if(bitfield[bi]&mask) {
	_pieceStats[index]->addCount();
      }
    }

  }
  std::sort(_sortedPieceStatIndexes.begin(), _sortedPieceStatIndexes.end(),
	    PieceStatRarer(_pieceStats));
}

void RarestPieceSelector::subtractPieceStats(const unsigned char* bitfield,
					     size_t bitfieldLength)
{
  size_t index = 0;
  for(size_t bi = 0; bi < bitfieldLength; ++bi) {
    
    for(size_t i = 0; i < 8; ++i, ++index) {
      unsigned char mask = 128 >> i;
      if(bitfield[bi]&mask) {
	_pieceStats[index]->subCount();
      }
    }

  }
  std::sort(_sortedPieceStatIndexes.begin(), _sortedPieceStatIndexes.end(),
	    PieceStatRarer(_pieceStats));
}

void RarestPieceSelector::updatePieceStats(const unsigned char* newBitfield,
					   size_t newBitfieldLength,
					   const unsigned char* oldBitfield)
{
  size_t index = 0;
  for(size_t bi = 0; bi < newBitfieldLength; ++bi) {
    
    for(size_t i = 0; i < 8; ++i, ++index) {
      unsigned char mask = 128 >> i;
      if((newBitfield[bi]&mask) && !(oldBitfield[bi]&mask)) {
	_pieceStats[index]->addCount();
      } else if(!(newBitfield[bi]&mask) && (oldBitfield[bi]&mask)) {
	_pieceStats[index]->subCount();
      }
    }

  }
  std::sort(_sortedPieceStatIndexes.begin(), _sortedPieceStatIndexes.end(),
	    PieceStatRarer(_pieceStats));
}

void RarestPieceSelector::addPieceStats(size_t index)
{
  std::vector<size_t>::iterator cur =
    std::lower_bound(_sortedPieceStatIndexes.begin(),
		     _sortedPieceStatIndexes.end(),
		     index, PieceStatRarer(_pieceStats));

  _pieceStats[index]->addCount();

  std::vector<size_t>::iterator to =
    std::upper_bound(cur+1, _sortedPieceStatIndexes.end(),
		     index, PieceStatRarer(_pieceStats));
  
  std::rotate(cur, cur+1, to);
}

const std::vector<size_t>&
RarestPieceSelector::getSortedPieceStatIndexes() const
{
  return _sortedPieceStatIndexes;
}

const std::vector<SharedHandle<PieceStat> >&
RarestPieceSelector::getPieceStats() const
{
  return _pieceStats;
}

} // namespace aria2
